#include <cstdlib>

#include "array.h"


namespace CSA
{


Array::Array(std::ifstream& file) :
  delete_array(true), item_index(0)
{
  file.read((char*)&(this->items), sizeof(this->items));
  file.read((char*)&(this->number_of_blocks), sizeof(this->number_of_blocks));
  file.read((char*)&(this->block_size), sizeof(this->block_size));

  this->array = new usint[this->block_size * this->number_of_blocks];
  file.read((char*)(this->array), this->block_size * this->number_of_blocks * sizeof(usint));
  this->buffer = new FastBitBuffer(this->array, this->block_size);

  this->integer_bits = length(this->items);
  this->samples = new FastBitBuffer(file, this->number_of_blocks + 1, this->integer_bits);

  this->buildIndex();
}

Array::Array(Array* original) :
  items(original->items),
  buffer(0), array(original->array), delete_array(false), block_size(original->block_size),
  number_of_blocks(original->number_of_blocks),
  samples(0), integer_bits(original->integer_bits),
  item_index(0), index_rate(original->index_rate)
{
  this->buffer = new FastBitBuffer(original->buffer);
  this->samples = new FastBitBuffer(original->samples);
  this->item_index = new FastBitBuffer(original->item_index);
}

Array::Array(ArrayEncoder& encoder) :
  items(encoder.items),
  delete_array(true),
  block_size(encoder.block_size),
  number_of_blocks(encoder.blocks),
  item_index(0)
{
  this->array = new usint[this->block_size * this->number_of_blocks];
  this->buffer = new FastBitBuffer(this->array, this->block_size);

  this->integer_bits = length(this->items);
  this->samples = new FastBitBuffer(this->number_of_blocks + 1, this->integer_bits);

  // Copy & linearize the array.
  usint pos = 0;
  for(std::list<usint*>::iterator iter = encoder.array_blocks.begin(); iter != encoder.array_blocks.end(); iter++)
  {
    memcpy(this->array + pos, *iter, encoder.superblock_bytes);
    pos += encoder.block_size * encoder.blocks_in_superblock;
  }
  memcpy(this->array + pos, encoder.array, encoder.current_blocks * encoder.block_size * sizeof(usint));

  // Compress the samples.
  for(std::list<usint*>::iterator iter = encoder.sample_blocks.begin(); iter != encoder.sample_blocks.end(); iter++)
  {
    usint* buf = *iter;
    for(usint i = 0; i < encoder.samples_in_superblock; i++)
    {
      this->samples->writeItem(buf[i]);
    }
  }
  for(usint i = 0; i < encoder.current_samples; i++)
  {
    this->samples->writeItem(encoder.samples[i]);
  }
  this->samples->writeItem(this->items);

  this->buildIndex();
}

Array::~Array()
{
  if(this->delete_array) { delete[] this->array; }
  delete   this->buffer;
  delete   this->samples;
  delete   this->item_index;
}

void
Array::writeTo(std::ofstream& file)
{
  file.write((char*)&(this->items), sizeof(this->items));
  file.write((char*)&(this->number_of_blocks), sizeof(this->number_of_blocks));
  file.write((char*)&(this->block_size), sizeof(this->block_size));
  file.write((char*)(this->array), this->block_size * this->number_of_blocks * sizeof(usint));
  this->samples->writeBuffer(file, false);
}

//--------------------------------------------------------------------------

usint
Array::getItem(usint index)
{
  if(index >= this->items) { return 0; }
  this->getSample(this->sampleForIndex(index));

  usint value = 0;
  while(this->sample + this->cur <= index)
  {
    value = this->buffer->readDeltaCode();
    this->cur++;
  }

  return value;
}

usint
Array::nextItem()
{
  if(this->cur >= this->block_items)
  {
    this->getSample(this->block + 1);
  }

  this->cur++;
  return this->buffer->readDeltaCode();
}

//--------------------------------------------------------------------------

usint
Array::reportSize()
{
  // We assume the reportSize() of derived classes includes any class variables of BitVector.
  usint bytes = this->buffer->reportSize();
  bytes += this->block_size * this->number_of_blocks * sizeof(usint);
  bytes += this->samples->reportSize();
  bytes += this->item_index->reportSize();
  return bytes;
}

//--------------------------------------------------------------------------

usint
Array::sampleForIndex(usint index)
{
  usint low = this->item_index->readItem(index / this->index_rate);
  usint high = this->item_index->readItem(index / this->index_rate + 1);

  this->samples->goToItem(low + 1);
  for(; low < high; low++)
  {
    if(this->samples->readItem() > index) { return low; }
  }

  return low;
}

//--------------------------------------------------------------------------

void
Array::buildIndex()
{
  delete this->item_index;

  usint index_samples = (this->number_of_blocks + Array::INDEX_RATE - 1) / Array::INDEX_RATE;
  this->index_rate = (this->items + index_samples - 1) / index_samples;
  index_samples = (this->items + this->index_rate - 1) / this->index_rate + 1;
  this->item_index = new FastBitBuffer(index_samples, length(this->number_of_blocks - 1));

  usint current = 0, pointer = 0;
  this->samples->goToItem(1);
  while(this->samples->hasNextItem())
  {
    usint limit = this->samples->readItem();
    while(current < limit)
    {
      this->item_index->writeItem(pointer);
      current += this->index_rate;
    }
    pointer++;
  }
  this->item_index->writeItem(this->number_of_blocks - 1);
}

//--------------------------------------------------------------------------

ArrayEncoder::ArrayEncoder(usint block_bytes, usint superblock_size) :
  items(0), blocks(1),
  block_size(BYTES_TO_WORDS(block_bytes)),
  superblock_bytes(superblock_size)
{
  this->array = new usint[this->superblock_bytes / sizeof(usint)];
  memset(this->array, 0, this->superblock_bytes);
  this->blocks_in_superblock = this->superblock_bytes / (sizeof(usint) * this->block_size);
  this->current_blocks = 1;

  this->samples = new usint[this->superblock_bytes / sizeof(usint)];
  this->samples_in_superblock = this->superblock_bytes / sizeof(usint);
  this->current_samples = 1;
  this->samples[0] = 0;

  this->buffer = new FastBitBuffer(this->array, this->block_size);
}

ArrayEncoder::~ArrayEncoder()
{
  delete[] this->array;

  delete this->buffer;
  for(std::list<usint*>::iterator iter = this->array_blocks.begin(); iter != this->array_blocks.end(); iter++)
  {
    delete[] *iter;
  }

  delete[] this->samples;
  for(std::list<usint*>::iterator iter = this->sample_blocks.begin(); iter != this->sample_blocks.end(); iter++)
  {
    delete[] *iter;
  }
}

void
ArrayEncoder::addItem(usint value)
{
  this->items++;
  if(this->buffer->writeDeltaCode(value)) { return; }

  // Didn't fit into the block. A new block & sample required.
  this->blocks++;
  this->current_blocks++;
  this->current_samples++;

  // Do we need a new superblock for the block?
  if(this->current_blocks > this->blocks_in_superblock)
  {
    this->array_blocks.push_back(this->array);
    this->array = new usint[this->superblock_bytes / sizeof(usint)];
    memset(this->array, 0, this->superblock_bytes);
    this->current_blocks = 1;
  }
  this->buffer->moveBuffer(this->array + (this->block_size * (this->current_blocks - 1)));

  // Do we need a new superblock for the sample?
  if(this->current_samples > this->samples_in_superblock)
  {
    this->sample_blocks.push_back(this->samples);
    this->samples = new usint[this->superblock_bytes / sizeof(usint)];
    this->current_samples = 1;
  }
  this->samples[this->current_samples - 1] = this->items - 1;

  // Finally, write the item.
  this->buffer->writeDeltaCode(value);
}


} // namespace CSA
