#ifndef ARRAY_H
#define ARRAY_H

#include <fstream>
#include <list>

#include "../misc/definitions.h"
#include "bitbuffer.h"


namespace CSA
{

/*
  Delta-coded array for storing positive integers.
  Do not try to encode value 0!
*/


class ArrayEncoder
{
  public:
    const static usint SUPERBLOCK_SIZE = MEGABYTE;

    // We assume superblock size is divisible by block and sample size.
    ArrayEncoder(usint block_bytes, usint superblock_size = SUPERBLOCK_SIZE);
    ~ArrayEncoder();

    void addItem(usint value);

    void addNewBlock();

    usint items, blocks;
    usint block_size, superblock_bytes;

    FastBitBuffer*   buffer;

    std::list<usint*> array_blocks;
    usint*            array;
    usint             blocks_in_superblock, current_blocks;

    std::list<usint*> sample_blocks;
    usint*            samples;
    usint             samples_in_superblock, current_samples;
};


class Array
{
  public:
    const static usint INDEX_RATE = 5;

    Array(std::ifstream& file);
    Array(Array* original);
    Array(ArrayEncoder& encoder);
    ~Array();

    void writeTo(std::ofstream& file);

//--------------------------------------------------------------------------

    // Returns 0 if the index is invalid.
    usint getItem(usint index);
    usint nextItem();

    inline bool hasNext()
    {
      return (this->sample + this->cur < this->items);
    }

//--------------------------------------------------------------------------

    inline usint getSize() { return this->items; }
    inline usint getBlockSize() { return this->block_size; }

    // This returns only the sizes of the dynamically allocated structures.
    usint reportSize();

  protected:
    usint          items;

    FastBitBuffer* buffer;
    usint*         array;
    bool           delete_array;  // Do we own the array?
    usint          block_size;
    usint          number_of_blocks;

    /*
      Each sample tells how many items are contained in the previous blocks.
    */
    FastBitBuffer* samples;
    usint          integer_bits;

    FastBitBuffer* item_index;
    usint          index_rate;

    // Iterator data.
    usint      block;
    usint      sample;
    usint      cur;
    usint      block_items;

    /*
      This function returns the sample corresponding to the block the given
      index might be found in. Parameter is assumed to be valid.
    */
    usint sampleForIndex(usint index);

    inline void getSample(usint sample_number)
    {
      this->block = sample_number;
      this->sample = this->samples->readItem(sample_number);
      this->cur = 0;
      this->block_items = this->samples->readItem() - this->sample;
      this->buffer->moveBuffer(this->array + (this->block * this->block_size));
    }

    /*
      Build a higher level index for faster queries.
      The index consists of about (number of samples) / INDEX_RATE pointers.
      The array cannot be used without the index.
    */
    void buildIndex();
};


} // namespace CSA


#endif // ARRAY_H
